/* Ari's JAVA implementation of a weighted, labelled graph structure specifically
	 for the minsum algorithm (i.e. optimized? for bipartite graphs */

/*  to work:  could have problems because enumerations do not guarantee an enumeration order...
		might conflict with the interfacing C++ program, but this is not clear to me now */

import java.applet.*;
import java.awt.*;
import java.util.*;
import java.io.*;

public class bipartiteGraph extends Frame implements Runnable {
	public Vector nodeList;       // the nodes in the graph
	public Vector edgeList;       // the edges in the graph
	public Vector messageList;    // a list of messages to place on the graph
	private Dimension size;       // maintins the size of the encompassing window
	private static boolean done=false;   // set true when the program should quit
	private DataInputStream input;// the input stream from where commands are issued

	// display constants
	private final int border=20;      // blank border on the outside of the graph
	private final int node_width=30;  // generic width of a node
	private final int node_height=30; // generic height of a node
	private final int PRETTY_ITERS=1000;  // the number of iterations for make_pretty
	private final int PRETTY_REFRESH=50;// # repeats/refresh
	private final int REFRESH_RATE=1000; // the number of milliseconds to wait, on minimum, between refreshes
	private final int MAX_PERTURB=1000;  // maximum number of perturbations in the heuristic edge-crossing reduction algorithm
	private final int SCREEN_BORDER=node_width;  // the desired border around the screen
	private final double RAND_THRESH=0.85;       // a threshold between 0 and 1 designed to keep things random...the higher the threshold, the less random certain choices will be
	private final Color DRAG_COLOR=Color.green; // the color of a dragged node

	// ... forces for makePretty(); lower numbers correspond to a greater force
	private final int SAME_CHECK_NODE_FORCE=13;  // two check nodes
	private final int SAME_SYMBOL_NODE_FORCE=15; // two symbol nodes
	private final int DIFFERENT_NODE_FORCE=18;   // one check node and one symbol node
	
	private final int EDGE_FORCE=1;  // one check node and one symbol node connected by an edge

	// display variables
	private float idealDist;         // the ideal distance between connected nodes (for makePretty)
	private final static int INCREMENT=10; // maximum amount to increment on each iteration of makePretty or minCrossings
	private boolean dontRepaint=false;

	// display variables
	private Node dragNode=null;      // the node being dragged

	// other variables
	private Random theRand;           // for pseudoRandom number generation
	private Thread paintThread;       // thread dedicated for repainting
	private Thread inputThread;       // thread dedicated for inputs
	private static boolean doRepaint; // indicates to the paintThread to repaint the screen
	private static boolean painting=false;  // has the painting thread been started?

	// UTILITIES
	private float force(float dist)
		/* %R:  dist can be positive or negative
			 %E:  computes a force of the same sign as dist, whose (absolute value)
			 magnitude is greatest at "ideal_dist"
			 */
	{
		float theForce;
		int theSign=MathUtils.sign(dist);
		float absDist=theSign*dist; // get actual distance (positive)
		float idealDiff=(idealDist-absDist);

		if (idealDiff==0) // we are ideal
			theForce=0;
		else
			theForce=MathUtils.square(idealDiff);
		
		// correct for sign
		if (Math.abs(dist)<idealDist)
			return -theSign*theForce; // then "go against the flow"
		else
			return theSign*theForce;  // "go with the flow"
	}
	
	// QUERY FUNCTIONS
	public boolean crossQ(Edge edge1, Edge edge2)
		/* %E:  returns true iff edges edge1 and edge2 cross */
	{
		// first check for a common node
		Node node1=edge1.start, node1p=edge1.end,
			node2=edge2.start, node2p=edge2.end, dummy;

		if (node1 == node2 ^
				node1 == node2p ^
				node1p   == node2 ^
				node1p   == node2p)
			return false;  // the edges share a common node
		
		// correct for orientation
		if (node1.left>node1p.left) {
			// switch them
			dummy=node1; node1=node1p; node1p=dummy;
		}
		if (node2.left>node2p.left) {
			// switch them
			dummy=node2; node2=node2p; node2p=dummy;
		}

		// check if they cross
		int x1=node1.left, y1=node1.upper,
			x2=node1p.left, y2=node1p.upper,
			x1p=node2.left, y1p=node2.upper,
			x2p=node2p.left, y2p=node2p.upper;
		if ((MathUtils.sign(y2-y2p)!=MathUtils.sign(y1-y1p)) &&
				(MathUtils.sign(x2-x1p)!=MathUtils.sign(x1-x2p)))
			return true;
		else
			return false;
	}

	public int numCrossings()
		/* %E:  returns the number of edge crossings in the current graph
		 */
	{
		int tempCrossings=0;                       // keeps track of the number of crossings so far

		try {
			for (int ii=0; ii<edgeList.size(); ii++) {
				Edge edge1=(Edge) edgeList.elementAt(ii);
				
				for (int jj=ii+1; jj<edgeList.size(); jj++) {
					Edge edge2=(Edge) edgeList.elementAt(jj);
					if (crossQ(edge1,edge2)) {
						// System.out.println("Edges "+edge1.name+" and "+edge2.name+" intersect.");
						tempCrossings++;                       // I've found one more crossing pair
					}
				}
			}
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("ERR:  Serious programming error in numCrossings; array out of bounds");
		}
		return tempCrossings;
	}


	// CONSTRUCTORS
	public void init() {
		nodeList=new Vector(1);
		edgeList=new Vector(0);
		messageList=new Vector(2); // should always start with one more items that there are nodes

		Node dummyNode=new Node(10,10,100,100);

		nodeList.addElement(dummyNode);
	}

	public void init(int left, int right)
		/* %R:  left, right>0
			 %E:  sets up nodes into a bipartite fashion, with "left"
			 left nodes and "right" right nodes
			 */
	{
		int ii;
		int vert_spacing,horiz_spacing;  // spacing on the window
		int node_no=0; // for debugging;

		theRand=new Random();            // initialize random numbers
		horiz_spacing=(size.width-2*border)/3;
		
		// System.out.println("Size(h)  :  "+size.height+"\n");
		// System.out.println("Size(w)  :  "+size.width +"\n");
		// System.out.println("Horizontal spacing:  "+horiz_spacing+"\n");
		// System.out.println("Vertical spacing  :  "+vert_spacing+"\n");

		// initialize nodeList and edgeList
		nodeList=new Vector(left+right);
		edgeList=new Vector(0);    // no edges initially
		messageList=new Vector(left+right+1); // messages for each node + 1 status message

		// add the left nodes
		vert_spacing=(size.height-2*border)/(left+1);
		for (ii=1;ii<=left;ii++) {
			Node dummyNode=new Node(horiz_spacing,
															ii*vert_spacing,
															node_width,
															node_height);
			dummyNode.setType(Node.symbolType); // designate as a symbol node
			dummyNode.changeName(String.valueOf(node_no++));
			//dummyNode.changeName("site "+String.valueOf(ii)); //assign the obvious labels
			nodeList.addElement(dummyNode);
		}
		
		// add the right nodes
		vert_spacing=(size.height-2*border)/(right+1);
		for (ii=1;ii<=right;ii++) {
			Node dummyNode=new Node(2*horiz_spacing,
															ii*vert_spacing,
															node_width,
															node_height);
			dummyNode.setType(Node.checkType); // designate as a check node
			dummyNode.changeName(String.valueOf(node_no++));
			//dummyNode.changeName("check "+String.valueOf(ii)); //assign the obvious labels
			nodeList.addElement(dummyNode);
		}

		// add the messages
		Message dummyMessage;
		for (ii=1; ii<=left+right; ii++) {
			// one for each node
			dummyMessage=new Message("",Color.blue,0,0);
			messageList.addElement(dummyMessage);
		}
		// one for the status message
		dummyMessage=new Message("",Color.red,0,0);
		messageList.addElement(dummyMessage);
	}

	public void changeNodeName(int nth, String newName)
		/* %R:  nth < |nodeList|
			 %M:  the n'th node in the nodeList
			 %E:  changes the name of the n'th node in the nodeList to newName
			 */
	{
		try {
			((Node) nodeList.elementAt(nth)).changeName(newName);
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("Array out of bounds in changeNodeName");
		}
	}

	public void changeEdgeName(int nth, String newName)
		/* %R:  nth < |edgeList|
			 %M:  the n'th edge in the edgeList
			 %E:  changes the name of the n'th edge in the edgeList to newName
			 */
	{
		try {
			((Edge) edgeList.elementAt(nth)).changeName(newName);
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("Serious Error:  Array out of bounds in changeEdgeName");
		}
	}

	public void changeNodeLoc(int nth, double xper, double yper)
		/* %R:  nth < |nodeList|
			 %M:  the n'th node in the nodeList; 0<=xper<=1.0; 0<=yper<=1.0
			 %E:  changes the upper-left location of the n'th node to screen coordinates (xper%,yper%),
			      where xper% is that percent of the screen size, and yper% is that percent of the screen size
			 */
	{
		try {
			Node theNode=(Node) nodeList.elementAt(nth);
			theNode.left=(int) (xper*size.width); theNode.upper=(int) (yper*size.height);
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("Array out of bounds in changeNodeLoc");
		}
	}

	public void addEdge(int node1, int node2)
		/* %R:  0 <= node1, node2 < |nodeList|
			 %E:  adds an edge between the node1'th and node2'th node
			 in the nodelist
			 */
	{
		Edge theEdge=new Edge((Node) nodeList.elementAt(node1),
													(Node) nodeList.elementAt(node2));
		theEdge.changeName(String.valueOf(edgeList.size())); // for debugging purposes
		edgeList.addElement(theEdge);
	}

 	// FOR PAINTING
	public void addMessage(String info, int left, int upper)
		/* %E:  adds a message "info" at displacements "left" and "upper"
		 */
	{
		Message theMessage=new Message(info,Color.blue,left,upper);
		messageList.addElement(theMessage);
	}

	public void changeMessage(int nth, String info, int left, int upper)
		/* %R:  nth < |messageList|
			 %M:  the n'th message in the messageList
			 %E:  makes the n'th message contain "info" at displacements "left" and "upper"
			 */
	{
		// System.out.println("Changing "+nth+"'th message to '"+info+"' at ("+left+","+upper+")");
		try {
			Message theMessage=(Message) messageList.elementAt(nth);
			theMessage.info=info;
			theMessage.leftDisp=left;
			theMessage.upperDisp=upper;
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("Array out of bounds in changeMessage");
		}
	}

	public void blockingRepaint() {
		/* 
			 %R:  should only be called from the inputThread
			 %E:  stall the input thread until we are done repainting
		 */
		doRepaint=true;
		while (doRepaint==true)
			inputThread.yield();
	}

	public void makePretty() {
		/* %M:  Node.left's and .upper's and .forcex's and .forcey's
			 repaints the screen occasionally for effect
			 %E:  attempts to move nodes around according to an atomic-force
			 model so that the graph looks better spaced out
			 */
		
		Enumeration allNodes, allEdges;
		idealDist=300; // (float) Math.sqrt((MathUtils.square(size.width)+MathUtils.square(size.height))/nodeList.size()); // roughly based on dividing the screen in squares
		
		System.out.println("Ideal distance between nodes is: "+idealDist);
		for (int repeats=0; repeats<PRETTY_ITERS; repeats++) {
			// reset forces to 0 initially
			allNodes=nodeList.elements();
			while (allNodes.hasMoreElements()) {
				Node theNode=(Node) allNodes.nextElement();
				theNode.forcex=(float) 0.0; theNode.forcey=(float) 0.0;
			}

			// go through each edge, computing related forces on the nodes
			// this way nodes with edges are more affected than nodes without edges
			allEdges=edgeList.elements();
			while (allEdges.hasMoreElements()) {
				Edge theEdge=(Edge) allEdges.nextElement();
				// compute respective forces
				float xdist=theEdge.end.left-theEdge.start.left;
				float ydist=theEdge.end.upper-theEdge.start.upper;
				// System.out.println("1: xdist is "+xdist+"; ydist is "+ydist);
				theEdge.start.forcex+=force(xdist)/EDGE_FORCE;
				theEdge.start.forcey+=force(ydist)/EDGE_FORCE;
				theEdge.end.forcex+=force(-xdist)/EDGE_FORCE;
				theEdge.end.forcey+=force(-ydist)/EDGE_FORCE;
			}

			// go through each node, computing a smaller related force on the nodes
			allNodes=nodeList.elements();
			while (allNodes.hasMoreElements()) {
				Node node1=(Node) allNodes.nextElement();
				Enumeration allNodes2=nodeList.elements();
				while (allNodes2.hasMoreElements()) {
					Node node2=(Node) allNodes2.nextElement();

					float xdist=node2.left-node1.left;
					float ydist=node2.upper-node1.upper;
					int factor; // the force factor between the two nodes
					if (node1.myType==Node.checkType && node2.myType==Node.checkType)
						factor=SAME_CHECK_NODE_FORCE;
					else if (node1.myType==Node.symbolType && node2.myType==Node.symbolType)
						factor=SAME_SYMBOL_NODE_FORCE;
					else
						factor=DIFFERENT_NODE_FORCE;

					// System.out.println("2: xdist is "+xdist+"; ydist is "+ydist);
					node1.forcex+=force(xdist)/factor;
					node1.forcey+=force(ydist)/factor;
					node2.forcex-=force(xdist)/factor;
					node2.forcey-=force(ydist)/factor;
				}
			
				// add strong forces from the sides of the screen if too close to the sides
				// if (node1.left<SCREEN_BORDER)
// 					node1.forcex+=nodeList.size()*force(0-node1.left); // force from the left
// 				if (node1.left>size.width-node1.width-SCREEN_BORDER)
// 					node1.forcex+=nodeList.size()*force(size.width-node1.left-node1.width); // force from the right
// 				if (node1.upper<SCREEN_BORDER)
// 					node1.forcey+=nodeList.size()*force(0-node1.upper); // force from the top
// 				if (node1.upper>size.height-node1.height-SCREEN_BORDER)
// 					node1.forcey+=nodeList.size()*force(size.height-node1.upper-node1.height); // force from the top
				// 1+ is to prevent division by 0
					node1.forcex+=
						MathUtils.square(size.width/(1+MathUtils.square(node1.left-0))); // force from left
					node1.forcex-=
						MathUtils.square(size.width/(1+MathUtils.square(node1.left-size.width))); // force from right
					node1.forcey+=
						MathUtils.square(size.width/(1+MathUtils.square(node1.upper-0))); // force from top
					node1.forcey-=
						MathUtils.square(size.width/(1+MathUtils.square(node1.upper-size.height))); // force from bottom
			}

			// see which node has the greatest magnitude force
			int debugCount=0;
			float max_force=-1; Node maxNode=(Node) nodeList.elementAt(0); // arbitrarily, the first node
			allNodes=nodeList.elements();
			while (allNodes.hasMoreElements()) {
				Node theNode=(Node) allNodes.nextElement();
				float total_force= (float)
					Math.sqrt(MathUtils.square(theNode.forcex)+MathUtils.square(theNode.forcey));
				//System.out.println("Force on Node "+(debugCount++)+": "+total_force+
				//									 "(x:"+theNode.forcex+ ", y:"+theNode.forcey+")");
				if (total_force>max_force && theRand.nextFloat()<RAND_THRESH) { // usually go by the maximum force (this way, if one node is stuck, then the other nodes will be moved)
					max_force=total_force;
					maxNode=theNode;
				}
			}

			// nudge the node appropriately
		if (theRand.nextFloat()<RAND_THRESH)
			// do the "right" thing
			if (Math.abs(maxNode.forcex)>Math.abs(maxNode.forcey))
				// nudge in the x direction
				maxNode.left+=((int) (theRand.nextFloat()*INCREMENT))*MathUtils.sign(maxNode.forcex);
			else
				// nudge in the y direction
				maxNode.upper+=((int) (theRand.nextFloat()*INCREMENT))*MathUtils.sign(maxNode.forcey);
		else
			// do the "wrong" thing
			if (Math.abs(maxNode.forcex)<=Math.abs(maxNode.forcey))
				// nudge in the x direction
				maxNode.left+=((int) (theRand.nextFloat()*INCREMENT))*MathUtils.sign(maxNode.forcex);
			else
				// nudge in the y direction
				maxNode.upper+=((int) (theRand.nextFloat()*INCREMENT))*MathUtils.sign(maxNode.forcey);


			// repaint
			if (repeats%PRETTY_REFRESH==0)
				blockingRepaint();
		}
	}

	private void swapCoords(Node node1, Node node2)
		/* %M:  node1, node2
			 %E:  swaps the coordinates of node1 and node2
			 */
	{
		Node dummy = new Node();   // temporarily holds node1's coordinates

		dummy.left=node1.left;
		dummy.upper=node1.upper;
		node1.left=node2.left;
		node1.upper=node2.upper;
		node2.left=dummy.left;
		node2.upper=dummy.upper;
	}

	public void minCrossings()
		/* %M:  node locations in the graph; repaints the screen occasionally
			 %E:  tries to heuristically reduce the number of edge crossings in the graph;
			 */
	{
		try {
			
			// record the number of crossings
			int xes=numCrossings();

			// FIRST try vertex (coordinate) swaps;
			System.out.println("Trying for vertex swaps...");
			for (int ii=0; ii<nodeList.size(); ii++)
				for (int jj=ii+1; jj<nodeList.size(); jj++) {
					// System.out.println("("+ii+","+jj+")");

					// try to switch the two nodes
					swapCoords((Node) nodeList.elementAt(ii),
										 (Node) nodeList.elementAt(jj));
					// see if the number of crossings has gone down
					int newXes=numCrossings();
					if (newXes<xes) {
						System.out.println("Number of crossings down to "+newXes);
						xes=newXes;
						// start at the beginning again
						ii=0; jj=1;
					}
					else { // revert to the previous coordinates
						swapCoords((Node) nodeList.elementAt(ii),
											 (Node) nodeList.elementAt(jj));
					}
				}
			
			// NEXT try random perturbation of vertices
			System.out.println("Trying for vertex perturbations...");
			xes=numCrossings();  // record the number of edge crossings
			int count=0;         // how many unsuccessful iterations we have seen
			for ( ;count<MAX_PERTURB && xes>0; count++) {
				// pick a random node
				Node randomNode=(Node) nodeList.elementAt((int) (theRand.nextFloat()*nodeList.size()));
				// pick a random perturbation
				int deltax=(int) (theRand.nextFloat()*INCREMENT)-(INCREMENT/2); // allow for negative numbers
				int deltay=(int) (theRand.nextFloat()*INCREMENT)-(INCREMENT/2); // allow for negative numbers

				// try perturbing the node
				randomNode.left+=deltax;
				randomNode.upper+=deltay;

				int newXes=numCrossings();
				if (newXes>xes) { // then swap back
					randomNode.left-=deltax;
					randomNode.upper-=deltay;
				}
				else {
					if (xes!=newXes)   // if we saw a real improvement
						System.out.println("Random perturbation improved crossings to: " + newXes);
					xes=newXes;
				}
			
				// repaint every so often for effect
				if (count%PRETTY_REFRESH==0)
					blockingRepaint();
			}
		}
		catch (ArrayIndexOutOfBoundsException e) {
			System.err.println("Serious Error:  Array out of bounds in changeEdgeName");
		}
	}

	//         PACKINGS
	public void squarePack()
		/* %R:  pack nodes in a square-like arrangement and then look
		**      for a plane isomorphism (ala minCrossings())
		*/
	{
		int ii,jj;
		int vert_spacing,horiz_spacing;  // spacing on the window

		Enumeration allNodes=nodeList.elements();
		horiz_spacing=(int) ((size.width-2*border)/(Math.sqrt(nodeList.size())+1)); // +1 for nicer spacing
		vert_spacing=(int) ((size.height-2*border)/(Math.sqrt(nodeList.size())+1));
		
		System.out.println("Horiz: "+horiz_spacing+", Vert :"+vert_spacing);
		for (ii=1,jj=1;allNodes.hasMoreElements();jj++) {
				if (jj>Math.sqrt(nodeList.size()))
					{ jj=1; ii++; }
				Node theNode= (Node) allNodes.nextElement();
				theNode.left=jj*horiz_spacing;
				theNode.upper=ii*vert_spacing;
			}

		// try to get rid of crossings
		minCrossings();
	}


	public void diamondPack()
		/* %R:  pack nodes in a square-like arrangement and then look
		**      for a plane isomorphism (ala minCrossings())
		*/
	{
		int ii,jj;
		int vert_spacing,horiz_spacing;  // spacing on the window

		Enumeration allNodes=nodeList.elements();
		horiz_spacing=(int) ((size.width-2*border)/(Math.sqrt(nodeList.size())+0.5)); // +1 for nicer spacing
		vert_spacing=(int) ((size.height-2*border)/(Math.sqrt(nodeList.size())+0.5));
		
		// System.out.println("Horiz: "+horiz_spacing+", Vert :"+vert_spacing);
		for (ii=1,jj=1;allNodes.hasMoreElements();jj++) {
				if (jj>((ii+1)%2)+Math.sqrt(nodeList.size()))
					{ jj=1; ii++; }
				Node theNode= (Node) allNodes.nextElement();
				theNode.left=jj*horiz_spacing-((ii+1)%2)*horiz_spacing/2;
				theNode.upper=ii*vert_spacing;
			}

		// try to get rid of crossings
		minCrossings();
	}

	// the actual painting procedure
	public void paint(Graphics g) {
		//System.out.println("Doing paint...");
		int count=0;

		// paint all the nodes
		Enumeration allNodes=nodeList.elements();
		while (allNodes.hasMoreElements()) {
			//System.out.println(" "+(count++)+" ");
			Node theNode=(Node) allNodes.nextElement();
			theNode.paint(g);
		}

		// paint all the nodes
		Enumeration allEdges=edgeList.elements();
		while (allEdges.hasMoreElements()) {
			//System.out.println(" "+(count++)+" ");
			Edge theEdge=(Edge) allEdges.nextElement();
			theEdge.paint(g);
		}

		// paint all the messages
		Enumeration allMessages=messageList.elements();
		while (allMessages.hasMoreElements()) {
			//System.out.println(" "+(count++)+" ");
			Message theMessage=(Message) allMessages.nextElement();
			theMessage.paint(g,size);
		}
	}

	public void repaint() {
		// repaints the screen
		Graphics g=getGraphics();
		size=getSize();  // reinitialize the size of the window
		g.clearRect(0,0,size.width,size.height);

		// fix up message nodes
		Node theNode;
		Message theMessage;
		if (messageList.size()>=nodeList.size()) {// else this makes no sense
				for (int ii=0; ii<nodeList.size(); ii++)
						try {
								theNode = (Node) nodeList.elementAt(ii);
								theMessage = (Message) messageList.elementAt(ii);
								
								theMessage.leftDisp=theNode.left+node_width/2-
										g.getFontMetrics().stringWidth(theMessage.info)/2;
								theMessage.upperDisp=theNode.upper+node_height+
										g.getFontMetrics().getHeight();
						}
						catch (ArrayIndexOutOfBoundsException theErr) {
								System.err.println("ERR:  Array out of bounds when redrawing node messages!");
						}
		}
		
		// actually paint
		paint(g);
	}
	
	public void start() {
		paintStart();
		inputStart();
	}

	public void paintStart() {
		//System.out.println("paint starting");

		// start up the paint thread
		if (paintThread==null) {
			//System.out.println("nulling");
			paintThread = new Thread(this);
			paintThread.start();
		}
	}

	public void inputStart() {
		//System.out.println("input starting");

		// start up the input thread
		if (inputThread==null) {
			//System.out.println("nulling");
			inputThread = new Thread(this);
			inputThread.start();
		}
	}

	public void run() {
		// paintThread.setPriority(Thread.NORM_PRIORITY);   // takes precedence
		if (!painting) {
			painting=true;
			paintRun();
		}
		else
			inputRun();
	}

	public void paintRun() {
		// set up the paint thread
		//System.out.println("paint running");
		while (!done) { // i.e. forever
			// System.out.println(" doRepaint: "+doRepaint);
			if (doRepaint==true) {// i.e. repaint the screen
				repaint();
				//System.out.println("repainting");
				doRepaint=false;    // finished repainting
			}
			paintThread.yield();
			// try {
			//		paintThread.sleep(REFRESH_RATE);   // sleep a little bit
			//	}
			//catch (InterruptedException e)
			//		{System.out.println("The paint thread was interrupted.");}
		}
	}

	public void inputRun() {
		//System.out.println("input running");
		// await input commands
		try {
			String line="foo";
			input=
				new DataInputStream(System.in); // make this into an input stream
			init();       // initial initialization

			for (;line!=null && !done;) {
				// System.out.print("> ");// the prompt
				System.out.print(">");
				while (input.available()==0)
					inputThread.yield(); // yield to the painting thread if not input is waiting
				line=input.readLine();
				ParseInput(line);
				if (!dontRepaint)
					blockingRepaint();
			}
		}
		catch (IOException e) {
			System.err.println("ERR:  Input stream problems...quitting...");
			end();
		}
	}

	public void end()
		/* %M:  done
			 %E:  signals termination */
	{
		done=true;
		//paintThread.sleep(100); // wait for input to finish
		//inputThread.sleep(100); // wait for paint to finish
		System.exit(0);         // exit the program
	}
			
	// INTERFACE

	private final void showHelp() {
		// prints out help information for the interface
		System.out.println(); // i.e. an empty line
		System.out.println("Interface options are: -> indicates a carriage return");
		System.out.println("  NOTE:  these commands affect only the interface");
		System.out.println("  init->a->b         initializes a bipartite graph with <a> left nodes");
		System.out.println("                     and <b> right nodes (if b=0 means regular graph)");
		System.out.println(); // i.e. an empty line
		
		System.out.println("  nodeName->n->name  the <n>'th node gets name <name>");
		System.out.println("  edgeName->n->name  the <n>'th edge gets name <name>");
		System.out.println("  moveNode->n->x->y  the <n>'th node is moved to coordiates (x%,y%)");
		System.out.println("                     of the screen size (0<=x,y<=1)");
		System.out.println("  addEdge->a->b      adds an edge from the <a>'th to <b>th nodes");
		System.out.println();

		System.out.println("  minCrossings       move nodes in order to heuristically minimize");
		System.out.println("                     the number of crossings");
		System.out.println("  numCrossings       returns the number of edge crossings in the graph");
		System.out.println();

		System.out.println("  makePretty         moves nodes to make the graph heuristically prettier");
		System.out.println("  squarePack         packs the nodes into a square lattice and tries to");
		System.out.println("                     eliminate crossings");
		System.out.println("  diamondPack        packs the nodes into a diamond lattice and tries to");
		System.out.println("                     eliminate crossings");
		System.out.println();

		System.out.println("  lightNode->n       highlights the <n>th node");
		System.out.println("  unlightNode->n     unhighlights the <n>th node");
		System.out.println();

		System.out.println("  messageNode->n->s  prints message <s> directly below the <n>th node");
		System.out.println("  changeInfo->n->s  changes text of the <n>th message to <s>");
		System.out.println("  lstatus->s         prints status message <s> on the bottom left of the screen");
		System.out.println("  rstatus->s         prints status message <s> on the bottom right of the screen");
		System.out.println();
			 
		System.out.println("  addMessage->s->a->b       (use with care!) adds a message <s> with (horiz,vert)");
		System.out.println("                            displacement =(a,b)");
		System.out.println("  changeMessage->n->s->a->b (use with care!) changes <n>'th message to <s>");
		System.out.println("                            at displacement (a,b)");
		System.out.println();

		System.out.println("  repaint            repaint the screen");
		System.out.println("  data               prints subsequent lines until another <data> string is found");
		System.out.println("  help               displays this help information");
		System.out.println("  quit               signals termination of the program");
	}

	public final void ParseInput(String line)
		/* %R:  line should be a non-null string
			 %M:  might read another line from the input if parameters are needed;
			 does this to avoid String parsing
			 %E:  runs the procedure dictated by line.  node/edge id's are according
			 to the order in which they were added to the graph

			 Available commands: (-> indicates a carraige return)
			 init->0           calls init()
			 init->a->b        (a!=0) calls init(a,b)

			 nodeName->a->name node "a" gets name "name"
			 edgeName->a->name edge "a" gets name "name"
			 moveNode->a->x->y node "a" moves to coordinates (x%,y%) of the screen (0<=x,y<=1)
			 addEdge->a->b     adds an edge from the a'th to b'th nodes in nodeList

			 minCrossings      move nodes around in order to heuristically minimize the number of crossings
			 numCrossings      returns the number of edge crossings in the graph

			 makePretty        moves nodes to make the graph prettier
			 squarePack        packs the nodes into a square lattice and tries to eliminate crossings
			 diamondPack       packs the nodes into a diamond lattice and tries to eliminate crossings

			 lightNode->a      highlights node "a"
			 unlightNode->a    unhighlights node "a"

			 messageNode->n->s prints message "s" directly below the n'th node
			 lstatus->s        prints status message "s" on the left hand side of the screen
			 rstatus->s        prints status message "s" on the right hand side of the screen
			 
			 // HANDLE WITH CARE!
			 addMessage->s->a->b    adds a message "s" with left displacement a and top displacement b;
			 changeMessage->n->s->a->b    changes message #n to "s" at displacements a and b
			 changeInfo->n->s  changes string of the n'th message to "s"

			 repaint           repaint the screen
			 data              prints subsequent lines until another "data" string is found
			 transdata         transparent "data"; doesn't print instructions
			 stopRepaint       stops repainting the screen without an explicit "repaint"; good for
			                   a long sequence of screen manipulations
			 startRepaint      (re)starts repainting the screen after each command (and does "repaint")
			 help              displays help information
			 quit              signals termination of the program
			 */
	{
		String nextLine;

		//System.out.println("Parse got "+line);

		try { 
			// init
			if (line.equalsIgnoreCase("init")) {
				nextLine=input.readLine();
				if (nextLine.equals("0"))
					init();
				else {
					System.out.println("extended init");
					int left=Integer.parseInt(nextLine);
					int right=Integer.parseInt(input.readLine());
					init(left,right);
				}
				doRepaint=true;
				System.out.println("inited");
			}
			
			// nodeName
			else if (line.equalsIgnoreCase("nodeName")) {
				int nth=Integer.parseInt(input.readLine());
				String newName=input.readLine();
				changeNodeName(nth,newName);
				// System.out.println("     [node: "+nth+"] is called "+newName+"\n");
			}

			// edgeName
			else if (line.equalsIgnoreCase("edgeName")) {
				int nth=Integer.parseInt(input.readLine());
				String newName=input.readLine();
				changeEdgeName(nth,newName);
				// System.out.println("     [edge: "+nth+"] is called "+newName+"\n");
			}
			
			// moveNode
			else if (line.equalsIgnoreCase("moveNode")) {
				int nth=Integer.parseInt(input.readLine());
				double x=Double.valueOf(input.readLine()).doubleValue();
				double y=Double.valueOf(input.readLine()).doubleValue();
				changeNodeLoc(nth,x,y);
				// System.out.println("     [edge: "+nth+"] is called "+newName+"\n");
			}

			// lightNode
			else if (line.equalsIgnoreCase("lightNode")) {
				int nth=Integer.parseInt(input.readLine());
				try {
					Node theNode=(Node) nodeList.elementAt(nth);
					theNode.sethighLight(true);
				}
				catch (ArrayIndexOutOfBoundsException e) {
					System.err.println("Array out of bounds while highlighting a node");
				}
			}

			// lightNode
			else if (line.equalsIgnoreCase("unlightNode")) {
				int nth=Integer.parseInt(input.readLine());
				try {
					Node theNode=(Node) nodeList.elementAt(nth);
					theNode.sethighLight(false);
				}
				catch (ArrayIndexOutOfBoundsException e) {
					System.err.println("Array out of bounds while highlighting a node");
				}
			}

			// addEdge
			else if (line.equalsIgnoreCase("addEdge")) {
				int a=Integer.parseInt(input.readLine());
				int b=Integer.parseInt(input.readLine());
				addEdge(a,b);
			}

			// addMessage
			// positive displacements are from the left (top) side of the screen;
			// negative displacements are from the right (bottom) side of the screen
			else if (line.equalsIgnoreCase("addMessage")) {
				String info=input.readLine();
				int a=Integer.parseInt(input.readLine());
				int b=Integer.parseInt(input.readLine());
				addMessage(info,a,b);
			}

			// changeMessage
			else if (line.equalsIgnoreCase("changeMessage")) {
				int nth=Integer.parseInt(input.readLine());
				String info=input.readLine();
				int a=Integer.parseInt(input.readLine());
				int b=Integer.parseInt(input.readLine());
				changeMessage(nth,info,a,b);
			}

			// changeInfo
			else if (line.equalsIgnoreCase("changeInfo")) {
				int nth=Integer.parseInt(input.readLine());
 				String info=input.readLine();

				// find the n'th message
				Message theMessage=null;
				try {
					theMessage=(Message) messageList.elementAt(nth);
				}
				catch (ArrayIndexOutOfBoundsException e) {
					System.err.println("Cannot find message for 'changeInfo'");
				}
				theMessage.info=info;
			}

			// messageNode
			else if (line.equalsIgnoreCase("messageNode")) {
				int nth=Integer.parseInt(input.readLine());
 				String info=input.readLine();

				// find the n'th node
				Node theNode=null;
				try {
					theNode=(Node) nodeList.elementAt(nth);
				}
				catch (ArrayIndexOutOfBoundsException e) {
					System.err.println("Cannot find node for 'messageNode'");
				}
				changeMessage(nth,
											info,
											theNode.left+node_width/2-getGraphics().getFontMetrics().stringWidth(info)/2,      // center around the node
											theNode.upper+node_height+getGraphics().getFontMetrics().getHeight() // position below the node
											);
			}

			// lstatus
			else if (line.equalsIgnoreCase("lstatus")) {
 				String info=input.readLine();
				changeMessage(nodeList.size(), // i.e. 1 after the last node message
											info,
											border,
											-getGraphics().getFontMetrics().getHeight()-border);
			}

			// rstatus
			else if (line.equalsIgnoreCase("rstatus")) {
 				String info=input.readLine();
				changeMessage(nodeList.size(), // i.e. 1 after the last node message
											info,
											-getGraphics().getFontMetrics().stringWidth(info)-border,
											-getGraphics().getFontMetrics().getHeight()-border);
			}

			// makePretty
			else if (line.equalsIgnoreCase("makePretty")) {
				makePretty();
			}

			// makePretty
			else if (line.equalsIgnoreCase("minCrossings")) {
				System.out.println("Starting with "+numCrossings()+" edge crossings.");
				minCrossings();
				System.out.println("Ending with   "+numCrossings()+" edge crossings.");
			}

			// numCrossings
			else if (line.equalsIgnoreCase("numCrossings")) {
				System.out.println("I see "+numCrossings()+" edge crossings in this graph.");
			}

			// squarePack
			else if (line.equalsIgnoreCase("squarePack")) {
				squarePack();
			}

			// diamondPack
			else if (line.equalsIgnoreCase("diamondPack")) {
				diamondPack();
			}

			// repaint
			else if (line.equalsIgnoreCase("repaint")) {
				doRepaint=true;
			}

			// data
			else if (line.equalsIgnoreCase("data")) {
				nextLine="data transmission; end with 'data\n'.";
				while (!nextLine.equalsIgnoreCase("data")) {
					System.out.println(nextLine);
					nextLine=input.readLine();
				}
			}

			// transdata
			else if (line.equalsIgnoreCase("transData")) {
				nextLine=input.readLine();
				while (!nextLine.equalsIgnoreCase("transData")) {
					System.out.println(nextLine);
					nextLine=input.readLine();
				}
			}

			// stopRepaint
			else if (line.equalsIgnoreCase("stopRepaint")) {
				dontRepaint=true;}

			// startRepaint
			else if (line.equalsIgnoreCase("startRepaint")) {
				dontRepaint=false;
				//repaint();
			}

			// help
			else if (line.equalsIgnoreCase("help"))
				showHelp();

			// quit
			else if (line.equalsIgnoreCase("quit"))
				end();
			
			else
				System.out.println(line+" was NOT managed\n");

		}
		catch (IOException e) {
			System.out.println("Input stream problems...quitting...");
			end();
		}
		catch(NumberFormatException n) {
			System.out.println("Number conversion problems...command aborted...");
		}
		//System.out.println("Done with parse "+line);
	}

	public boolean mouseDown(Event e, int xx, int yy) { // pick a node to drag
		// find the node
		//System.out.println("Mouse is down");
		Enumeration allNodes=nodeList.elements();
		boolean found=false; // has the node been found
		Node theNode=null;   // the node we have found

		while (allNodes.hasMoreElements() && !found) {
			//System.out.println(" "+(count++)+" ");
			theNode=(Node) allNodes.nextElement();
			// compute the distance to the node from the mouse click
			float dist=MathUtils.square(xx-(theNode.left+theNode.width/2))+
				MathUtils.square(yy-(theNode.upper+theNode.height/2));
			if (dist<=MathUtils.square(theNode.width)+MathUtils.square(theNode.height)+1) // +1 for some lee-way
				found=true; // we have found theNode
		}

		if (found) {
			// mark the node for dragging
			dragNode=theNode;
			// highlight the node
			Graphics g=getGraphics();
			g.setXORMode(DRAG_COLOR);
			g.fillOval(theNode.left,theNode.upper,theNode.width,theNode.height);
			g.setPaintMode();
		}
		else { // no node to drag
			dragNode=null;
		}

		paintThread.yield(); // yield control to input text as necessary
		return true;
	}

	public boolean mouseDrag(Event e, int x, int y) {
		if (dragNode==null)
			return true; // i.e. nothing to drag
		else {// remove the old node
			Graphics g=getGraphics();
			g.setXORMode(DRAG_COLOR);
			g.fillOval(dragNode.left,dragNode.upper,dragNode.width,dragNode.height);
			dragNode.left=x;
			dragNode.upper=y;
			g.fillOval(dragNode.left,dragNode.upper,dragNode.width,dragNode.height);
			g.setPaintMode();
		}

		paintThread.yield(); // yield control to input text as necessary
		return true;

	}
	
	// release the node if necessary
	public boolean mouseUp(Event e, int xx, int yy) {
		if (dragNode==null)
			return true; // i.e. nothing to release
		else {
			// fix up the message location
			Message theMessage;
			try {
				theMessage = (Message) messageList.elementAt(nodeList.indexOf(dragNode));
			theMessage.leftDisp=dragNode.left+node_width/2-
				getGraphics().getFontMetrics().stringWidth(theMessage.info)/2;
			theMessage.upperDisp=dragNode.upper+node_height+
				getGraphics().getFontMetrics().getHeight();
			}
			catch (ArrayIndexOutOfBoundsException theErr) {
				System.err.println("ERR:  Cannot fix message of moved node!");
			}

			// fix up the data structures
			dragNode=null;  // i.e. nothing to drag any more
			doRepaint=true; // request a repaint
		}

		paintThread.yield(); // yield control to input text as necessary
		return true;
	}

	public static void main(String argv[]) { // do nothing for now
		bipartiteGraph myGraph = new bipartiteGraph();

		// "display" info
		myGraph.setBackground(Color.white);
		myGraph.setForeground(Color.black);
		myGraph.setSize(800,800);
		myGraph.size=myGraph.getSize();
		myGraph.setResizable(true);
		myGraph.show();
		myGraph.start(); // start up painting thread
		
		System.out.println("Welcome to the (bipartite) graph interface by Ari Trachtenberg (arit@alum.mit.edu).");
		System.out.println("Feel free to change this program source in any way, as long as I am");
		System.out.println("referenced, and your name as a modifier is clearly stated.");
		myGraph.showHelp();
		//while (!myGraph.done) { // wiat for the all done
		//	yield()} // all done
		//System.exit(0);
	}
}
